'use strict';

const EventEmitter = require('events');
const amqplib      = require('amqplib');
const Promise      = require('bluebird');
const domain       = require('domain');

class EventBus {

    constructor() {
        this.initialized = false;
    }

    _ensureEventbusInitializaed() {
        if (!this.initialized) {
            throw new Error('Eventbus was not initialized');
        }
    }

    init(url, microServiceName) {
        if (!microServiceName) {
            throw new Error('Micro service name missing');
        }
        this.microServiceName = microServiceName;
        this.openPromise      = amqplib.connect(url);
        this.publishChannel;
        this.initialized = true;
    }

    publish(eventName, msg) {
        this._ensureEventbusInitializaed();

        return Promise.resolve()
            .then(() => {
                if (!this.publishChannel) {
                    return this._createPublishChannel();
                }
            })
            .then(() => {
                return this.publishChannel.assertExchange(eventName, 'fanout', { durable: true });
            })
            .then(() => {
                return this.publishChannel.publish(eventName, 'my-routing-key', new Buffer(this._serialize(msg)));
            });
    }

    subscribe(eventName, handler) {
        this._ensureEventbusInitializaed();
        const subscriberEmitter = new EventEmitter();

        const queueName = `${this.microServiceName}_${eventName}`;
        Promise.resolve()
            .then(() => {
                return this._getSubscribeChannel();
            })
            .then((ch) => {
                return ch.assertExchange(eventName, 'fanout', { durable: true })
                    .then(() => {
                        return ch.assertQueue(queueName, { durable: true })
                            .then(() => {
                                return ch.bindQueue(queueName, eventName);
                            })
                            .then(() => {
                                return ch.consume(queueName, (msg) => {
                                    const d = domain.create();
                                    d.on('error', (err) => {
                                        subscriberEmitter.emit('error', { handler: err });
                                    });
                                    d.run(() => {
                                        ch.ack(msg); // TODO allow controlling when to ack
                                        handler(this._deserialize(msg.content.toString()))
                                            .catch((err) => {
                                                subscriberEmitter.emit('error', { handler: err });
                                            })
                                            .done();
                                    });
                                });
                            });
                    });
            })
            .catch((err) => {
                subscriberEmitter.emit('error', { connection: err });
            })
            .done();

        return subscriberEmitter;
    }

    _createPublishChannel() {
        return this.openPromise
            .then((conn) => {
                return conn.createChannel();
            })
            .then((ch) => {
                this.publishChannel = ch;
            });
    }

    _getSubscribeChannel() {
        return this.openPromise
            .then((conn) => {
                return conn.createChannel();
            });
    }

    _serialize(body) {
        return JSON.stringify(body);
    }

    _deserialize(content) {
        return JSON.parse(content);
    }

}

module.exports = new EventBus();
